package lsgui.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Paint;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import javax.swing.JPanel;
import lsgui.jna.data.ShooterTGAImage;
import lsgui.util.LaterationImage;
import lsgui.util.LaterationImageListener;

public class Canvas extends JPanel {

    private boolean eventsOn = true;
    private boolean statisticsOn = false;
    private DecimalFormat format = new DecimalFormat("0.00");

    /** Offscreen image for drawing */
    private Image image;

    /** Graphics object of offscreen image */
    private Graphics2D buffer;

    /** The calculated image */
    private LaterationImage latImage;
    
    /** Width and height of image and therefore canvas */
    private int width, height;

    private final ArrayList<java.awt.Point> anchors;
    
    private List<LaterationImageListener> listeners;

    public static final int MAX_A = 15;

    public Canvas(int width, int height) {
        this.width = width;
        this.height = height;
        anchors = new ArrayList<java.awt.Point>();
        listeners = new ArrayList<LaterationImageListener>();
        setPreferredSize(new Dimension(width, height));
        addMouseListener(new MouseListener() {

            public void mouseClicked(MouseEvent me) {
                if (!eventsOn) {
                    return;
                }
                if (me.getButton() == MouseEvent.BUTTON1) {
                    if (me.getX() > Canvas.this.width) {
                        return;
                    }
                    if (me.getY() > Canvas.this.height) {
                        return;
                    }
                    if (anchors.size() >= MAX_A) {
                        return;
                    }
                    anchors.add(me.getPoint());
                    repaintRequest();
                }
                if (me.getButton() == MouseEvent.BUTTON3) {
                    // Test if clicked on an anchor node
                    for (int i = anchors.size()-1; i >= 0; i--) {
                        java.awt.geom.Ellipse2D.Double circle =
                                new java.awt.geom.Ellipse2D.Double(
                                anchors.get(i).x - 5,
                                anchors.get(i).y - 5,
                                10, 10);
                        if (circle.contains(me.getX(), me.getY())) {
                            anchors.remove(i);
                            repaintRequest();
                            return;
                        }
                    }
                }
            }

            public void mousePressed(MouseEvent me) {

            }

            public void mouseReleased(MouseEvent me) {

            }

            public void mouseEntered(MouseEvent me) {

            }

            public void mouseExited(MouseEvent me) {

            }

        });
    }

    public int getCanvasWidth() {
        return width;
    }
    
    public int getCanvasHeight() {
        return height;
    }
    
    public void addLaterationImageListener(LaterationImageListener l) {
        if (!listeners.contains(l)) {
            listeners.add(l);
        }
    }
    
    private void notifyLaterationImageListeners() {
        for (LaterationImageListener l : listeners) {
            l.imageChanged();
        }
    }
    
    public void setStatisticsEnabled(boolean enabled) {
        statisticsOn = enabled;
        repaintRequest();
    }
    
    @Override
    protected void paintComponent(Graphics g) {
        if (image == null) {
            repaintOffScreenImage();
        }
        g.drawImage(image, 0, 0, this);
    }

    private void repaintOffScreenImage() {
        if (image == null || (! (image.getWidth(this) == getSize().width
                && image.getHeight(this) == getSize().height))) {
            image = createImage(getSize().width, getSize().height);
            if (image == null) {
                return;
            }
            buffer = (Graphics2D) image.getGraphics();
            buffer.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                             RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        }
        
        super.paintComponent(buffer);
        Paint oldPaint = buffer.getPaint();
        Font oldFont = buffer.getFont();

        if (latImage != null && latImage.hasImages()) {
            LaterationImage.RenderedImage rImage = latImage.getRenderedImage();
            buffer.drawImage(rImage.image, 0, 0, null);
        } else {
            buffer.setPaint(Color.LIGHT_GRAY);
            buffer.fillRect(0, 0, this.width, this.height);
        }

        // Draw anchor nodes
        drawAnchors(buffer);

        // Draw statistics
        if (statisticsOn && latImage != null && latImage.hasImages() &&
                !latImage.isCompareImage()) {
            LaterationImage.RenderedImage rImage = latImage.getRenderedImage();
            paintHistogram(buffer, 0, 0, rImage.percentage,
                    rImage.average, rImage.positionErrors,
                    rImage.histogram, rImage.getCutoff());
        }

        buffer.setPaint(oldPaint);
        buffer.setFont(oldFont);
    }
    
    public void drawAnchors(Graphics2D buffer) {
        int i = 0;
        buffer.setPaint(Color.RED);
        synchronized (anchors) {
            Iterator<java.awt.Point> iter = anchors.iterator();
            while (MainFrame.drawAnchors && iter.hasNext()) {
                java.awt.Point p = iter.next();
                buffer.fillOval((int)(p.x)-3, (int)(p.y)-3, 6, 6);
                buffer.drawString(""+(i+1), (int) (p.x) - 4, (int) (p.y) - 5);
                i++;
            }
        }
    }
    
    private void paintHistogram(Graphics2D g, int xPos, int yPos, int[] percentage, double avg, int calcFail, int[] histogram, double cutoff) {
        int statW = 300;
        int statH = 450;
        int borderL = 50 + xPos;
        int borderT = 50 + yPos;
        g.setFont(new Font("Courier", Font.PLAIN, 12));
        int fHeight = buffer.getFontMetrics().getHeight();
        g.setPaint(Color.WHITE);
        g.fillRect(borderL, borderT, statW, statH);
        g.setPaint(Color.BLACK);
        g.drawString("Statistics:", borderL + 5, borderT + fHeight + 5);
        int[] percent = new int[] {50, 75, 90};
        for (int j = percentage.length-1; j >= 0; j--) {
            g.drawString(percent[j] + "% <= " + percentage[j] + "% of MAX_RANGE_ERROR", borderL + 5,
                    borderT + fHeight + 5 + 2 * fHeight + j * fHeight);
        }
        g.drawString("Average error: " + format.format(avg), borderL + 5,
                    borderT + fHeight + 5 + 2 * fHeight + 4 * fHeight);
        g.drawString("Position errors: " + calcFail, borderL + 5,
                    borderT + fHeight + 5 + 2 * fHeight + 5 * fHeight);

        int xs = borderL + 10 + 1;
        int ys = borderT + fHeight + 5 + 3 * fHeight + 7 * fHeight;
        g.setFont(new Font("Courier", Font.PLAIN, 10));
        fHeight = g.getFontMetrics().getHeight();
        g.drawLine(borderL + 10, ys, borderL + 10, statH + borderT - 5);
        g.drawLine(borderL + 5, statH + borderT - 10,
                borderL + statW - 5, statH + borderT - 10);
        ys += 10;

        // draw red line at top (cutoff line)
        g.setPaint(Color.RED);
        g.drawLine(xs, ys, histogram.length + 4 + xs, ys);
        g.drawString("40%", borderL + statW - 25, ys + fHeight/2);
        int plotH = (statH + borderT - 11) - ys;
        g.setPaint(Color.LIGHT_GRAY);
        g.drawLine(xs, ys + plotH/4, histogram.length + 4 + xs, ys + plotH/4);
        g.drawString("30%", borderL + statW - 25, ys + plotH/4 + fHeight/2);
        g.drawLine(xs, ys + plotH/2, histogram.length + 4 + xs, ys + plotH/2);
        g.drawString("20%", borderL + statW - 25, ys + plotH/2 + fHeight/2);
        g.drawLine(xs, ys + 3*plotH/4, histogram.length + 4 + xs, ys + 3*plotH/4);
        g.drawString("10%", borderL + statW - 25, ys + 3*plotH/4 + fHeight/2);

        // draw color map below x-axis
        byte[] colorMap = ShooterTGAImage.buildColorMap(ShooterTGAImage.VERY_GOOD_VALUES_COLOR);
        for (int j = 0; j < 250; j++) {
            g.setPaint(new Color(ShooterTGAImage.btoi(colorMap[3*j+2]),
                    ShooterTGAImage.btoi(colorMap[3*j+1]), ShooterTGAImage.btoi(colorMap[3*j])));
            g.drawLine(xs + j, statH + borderT - 9,
                xs + j, statH + borderT - 4);
        }
        for (int j = 250; j < 260; j++) {
            g.setPaint(new Color(ShooterTGAImage.btoi(colorMap[3*250+2]),
                    ShooterTGAImage.btoi(colorMap[3*250+1]), ShooterTGAImage.btoi(colorMap[3*250])));
            g.drawLine(xs + j, statH + borderT - 9,
                xs + j, statH + borderT - 4);
        }

        // draw buckets,  10 pixel each (1m)
        ys = (statH + borderT - 11) - ys;
        g.setPaint(Color.GRAY);
        int sum = 0;
        for (int j = 0; j < histogram.length; j++) {
            sum += histogram[j];
            if ((j+1) % 10 == 0) {
                int tmp = (int) ((sum / cutoff) * ys);
                if (tmp > ys) {
                    tmp = ys;
                }
                if (tmp >= 1) {
                    g.setPaint(new Color(ShooterTGAImage.btoi(colorMap[3*j+2]),
                        ShooterTGAImage.btoi(colorMap[3*j+1]), ShooterTGAImage.btoi(colorMap[3*j])));
                    g.fillRect(xs + j - 9, statH + borderT - 11 - tmp, 10, tmp+1);
                }
                sum = 0;
            }
            if (j == 250) {
                // draw blue bucket at end
                int tmp = (int) ((histogram[250] / cutoff) * ys);
                if (tmp > ys) {
                    tmp = ys;
                }
                if (tmp >= 1) {
                    g.setPaint(new Color(ShooterTGAImage.btoi(colorMap[3*250+2]),
                        ShooterTGAImage.btoi(colorMap[3*250+1]), ShooterTGAImage.btoi(colorMap[3*250])));
                    g.fillRect(xs + 250, statH + borderT - 11 - tmp, 10, tmp+1);
                }
                break;
            }
        }
    }

    public Image getOffScreenImage() {
        return image;
    }

    public ArrayList<java.awt.Point> getAnchors() {
        return anchors;
    }

    public void setEventsOn(boolean b) {
        eventsOn = b;
    }

    public LaterationImage getLatImage() {
        return latImage;
    }

    public void setLatImage(LaterationImage image) {
        boolean adjust = false;
        if (this.width != image.getImageWidth()) {
            this.width = image.getImageWidth();
            adjust = true;
        }
        if (this.height != image.getImageHeight()) {
            this.height = image.getImageHeight();
            adjust = true;
        }
        if (adjust) {
            setSize(width, height);
            setPreferredSize(new Dimension(width, height));
        }
        latImage = image;
        notifyLaterationImageListeners();
    }

    public boolean checkAnchors(List<Point> anchors) {
        for (Point p : anchors) {
            if (p.x < 0 || p.x >= this.width || p.y < 0 || p.y >= this.height) {
                return false;
            }
        }
        return true;
    }
    
    public void setAnchorsWOR(Point[] anchors) {
        synchronized (anchors) {
            this.anchors.clear();
            this.anchors.addAll(Arrays.asList(anchors));
        }
    }

    public void setAnchors(Point[] anchors) {
        setAnchors(Arrays.asList(anchors));
    }
    
    public void setAnchors(List<Point> anchors) {
        synchronized (anchors) {
            this.anchors.clear();
            this.anchors.addAll(anchors);
        }
        repaintRequest();
    }

    public void reset() {
        latImage = null;
        anchors.clear();
        notifyLaterationImageListeners();
    }

    public void repaintRequest() {
        repaintOffScreenImage();
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                revalidate();
                repaint();
            }
        });
    }

}
