package lsgui.util;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import lsgui.Main;
import lsgui.jna.data.LS2Result;
import lsgui.jna.data.ShooterTGAImage;
import lsgui.verify.OutputVariant;

public class LaterationImage {

    private int engine;
    private int imageWidth;
    private int imageHeight;
    private int imageIndex = -1;
    private int imageMode;
    private List<LaterationImageData> images;
    
    /** Algorithm information */
    private int algorithmType;
    private String algorithmName;
    
    /** Current rendered image to display */
    private RenderedImage renderedImage;
    
    public static final int TYPE_LOCATION_ALGORITHM = 1;
    public static final int TYPE_LOWER_BOUND_ESTIMATOR = 2;
    
    public static final int ENGINE_LIBRARY = 1;
    public static final int ENGINE_VERIFY = 2;
    
    public static class LaterationImageData {
        
        private String tag;
        private String fileTag;
        private float[] imageData;

        public LaterationImageData(float[] imageData, OutputVariant variant) {
            this.imageData = imageData;
            this.tag = variant.tag();
            this.fileTag = variant.fileTag();
        }
        
        public LaterationImageData(float[] imageData, String variant) {
            this.imageData = imageData;
            this.tag = variant;
            this.fileTag = variant.replaceAll(" ", "_");
        }

        public float[] getImageData() {
            return imageData;
        }

        public String getTag() {
            return tag;
        }

        public String getFileTag() {
            return fileTag;
        }

    }
    
    public static class RenderedImage {
        public double average;
        public int[] histogram;
        public int[] percentage;
        public int positionErrors;
        public BufferedImage image;
     
        private int total;
        private int[] percent;
        
        public RenderedImage(int total) {
            this.total = total;
            this.percent = new int[4];
            this.percent[0] = (50 * this.total) / 100;
            this.percent[1] = (75 * this.total) / 100;
            this.percent[2] = (90 * this.total) / 100;
            this.percent[3] = total + 1;
            this.histogram = new int[255];
            this.positionErrors = 0;
        }
        
        public int getCutoff() {
            return (40 * total) / 100;
        }
        
        private void calculateStatistics() {
            average = 0;
            percentage = new int[3];
            int idx = 0, sum = 0;
            for (int i = 0; i <= 250; i++) {
                sum += histogram[i];
                average += histogram[i] * i;
                if (sum > percent[idx]) {
                    percentage[idx++] = i;
                }
            }
            average /= total;
        }
    }
    
    public LaterationImage(LS2Result result,
                           int engine,
                           int imageMode) {
        this.engine = engine;
        this.imageWidth = result.width;
        this.imageHeight = result.height;
        this.imageMode = imageMode;
        this.images = new ArrayList<LaterationImageData>();
        for (int i = 0; i < result.data.length; i++) {
            if (result.data[i] != null) {
                LaterationImageData lid;
                if (engine == ENGINE_VERIFY) {
                    lid = new LaterationImageData(result.data[i],
                            OutputVariant.values()[i]);
                } else {
                    lid = new LaterationImageData(result.data[i],
                            Main.outputVariants[i]);
                }
                this.images.add(lid);
            }
        }
        if (images.size() > 0) {
            setActiveImage0(0);
        }
    }

    public boolean isCompareImage() {
        return imageMode != ShooterTGAImage.IMAGE_MODE_ONE;
    }
    
    public int getEngine() {
        return engine;
    }

    public int getAlgorithmType() {
        return algorithmType;
    }

    public void setAlgorithmType(int algorithmType) {
        this.algorithmType = algorithmType;
    }

    public String getAlgorithmName() {
        return algorithmName;
    }

    public void setAlgorithmName(String algorithmName) {
        this.algorithmName = algorithmName;
    }

    public int getImageWidth() {
        return imageWidth;
    }

    public int getImageHeight() {
        return imageHeight;
    }

    public int getImageIndex() {
        return imageIndex;
    }
    
    public boolean hasImages() {
        return getImageCount() > 0;
    }
    
    public int getImageCount() {
        return this.images.size();
    }
    
    public LaterationImageData getLaterationImageData() {
        return imageIndex != -1 ? this.images.get(imageIndex) : null;
    }
    
    public LaterationImageData getLaterationImageData(int index) {
        if (index >= 0 && index < images.size()) {
            return this.images.get(index);
        }
        return null;
    }

    public RenderedImage getRenderedImage() {
        return renderedImage;
    }
    
    public void setActiveImage(int index) {
        if (index >= 0 && index < images.size()) {
            setActiveImage0(index);
        }
    }
    
    private void setActiveImage0(int index) {
        int oldImageIndex = this.imageIndex;
        try {
            this.imageIndex = index;
            ShooterTGAImage imgTga = new ShooterTGAImage(this, imageMode);
            renderedImage = new RenderedImage(imageWidth * imageHeight);
            imgTga.decode(renderedImage);
            renderedImage.calculateStatistics();
        } catch (IOException e) {
            this.imageIndex = oldImageIndex;
        }
    }
    
    public void firstImage() {
        setActiveImage(0);
    }
    
    public void nextImage() {
        int next = Math.min(++imageIndex, getImageCount()-1);
        setActiveImage(next);
    }
    
    public void previousImage() {
        int next = Math.max(--imageIndex, 0);
        setActiveImage(next);
    }
    
    public void lastImage() {
        setActiveImage(getImageCount()-1);
    }
    
}
