package lsgui.verify;

import latmath.algorithm.BasicLaterationAlgorithm;
import latmath.algorithm.LowerBoundEstimator;
import latmath.errormodel.ErrorModel;
import latmath.util.Point2d;
import lsgui.jna.data.LS2Result;

public class Verifier implements Runnable {

    private int offset;
    private int length;
    private LS2Result result;
    private Point2d[] points;
    private Thread owner;
    
    private ImagePositionEvaluator imgEvaluator;

    private volatile int currentStep = -1;
    private volatile boolean stop;

    public Verifier(int runs,
                    int offset,
                    int length,
                    LS2Result result,
                    ErrorModel errorModel,
                    BasicLaterationAlgorithm alg,
                    Point2d[] points,
                    Thread owner) {
        this.offset = offset;
        this.length = length;
        this.result = result;
        this.points = points;
        this.owner = owner;
        if (alg instanceof LowerBoundEstimator) {
            imgEvaluator = new SingleLowerBoundEstimatorEvaluator(alg, points, errorModel);
        } else {
            imgEvaluator = new SingleAlgorithmEvaluator(runs, alg, points, errorModel);
        }
        stop = false;
    }
    
    public int getCurrentStep() {
        return currentStep + 1;
    }

    public void stop() {
        stop = true;
    }

    public void run() {
        int lineIdx, currentIdx;
        Point2d current = new Point2d(0, 0);
        double[] realDist = new double[points.length];
        for (int i = offset; i < offset+length; i++) {
            lineIdx = i * result.width;
            for (int j = 0; j < result.width; j++) {
                currentIdx = lineIdx + j;
                if (stop) {
                    // fast abort, only invalidate rest of image not calculated
                    for (int k = 0; k < result.data.length; k++) {
                        if (result.data[k] != null) {
                            result.data[k][currentIdx] = LS2Result.ERROR;
                        }
                    }
                    continue;
                }
                current.x = j;
                current.y = i;
                for (int l = 0; l < points.length; l++) {
                    realDist[l] = current.distance(points[l]);
                }
                // call image evaluator with given data
                imgEvaluator.evaluatePosition(current, realDist, currentIdx, result);
            }
            currentStep = i - offset;
        }
        owner.interrupt();
    }

    private interface ImagePositionEvaluator {
        /**
         * Evaluate image(s) at given location with an algorithm evaluator.
         * 
         * @param position The current position.
         * @param distances The real distances to the anchor nodes.
         * @param index The result index of the image(s).
         * @param result The result data of the image(s).
         */
        void evaluatePosition(Point2d position, double[] distances, int index,
                LS2Result result);
    }
    
    private static class SingleAlgorithmEvaluator implements ImagePositionEvaluator {

        private int runs;
        private Point2d[] anchors;
        private ErrorModel errorModel;
        private double[] erroneousDistances;
        private BasicLaterationAlgorithm algorithm;
        
        public SingleAlgorithmEvaluator(int runs, BasicLaterationAlgorithm algorithm,
                Point2d[] anchors, ErrorModel errorModel) {
            this.runs = runs;
            this.anchors = anchors;
            this.errorModel = errorModel;
            this.erroneousDistances = new double[anchors.length];
            this.algorithm = algorithm;
        }

        public void evaluatePosition(Point2d position, double[] distances,
                int index, LS2Result result) {
            int count = 0;
            double error = 0, squaredError = 0, minError = Double.MAX_VALUE, maxError = Double.MIN_VALUE;
            for (int k = 0; k < runs; k++) {
                for (int l = 0; l < anchors.length; l++) {
                    erroneousDistances[l] = distances[l] + errorModel.getOffset(distances[l], 0);
                }
                Point2d estimate = algorithm.localize(anchors, erroneousDistances, position, errorModel, result.width, result.height);
                // Stop on first estimation error, overall result for this position failed
                if (estimate == null) {
                    count = 0;
                    break;
                }
                count++;
                double tmp = estimate.distance(position);
                error += tmp;
                squaredError += tmp * tmp;
                if (tmp > maxError) {
                    maxError = tmp;
                }
                if (tmp < minError) {
                    minError = tmp;
                }
            }
            double mse = count > 0 ? squaredError / count : 0;
            // Write results back at given position
            if (result.data[OutputVariant.AVERAGE_ERROR.ordinal()] != null) {
                result.data[OutputVariant.AVERAGE_ERROR.ordinal()][index] = applyValue(count, 1, (float)error);
            }
            if (result.data[OutputVariant.MAXIMUM_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MAXIMUM_ERROR.ordinal()][index] = applyValue(count > 0 ? 1 : 0, 1, (float)maxError);
            }
            if (result.data[OutputVariant.MINIMUM_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MINIMUM_ERROR.ordinal()][index] = applyValue(count > 0 ? 1 : 0, 1, (float)minError);
            }
            if (result.data[OutputVariant.MEAN_SQUARED_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MEAN_SQUARED_ERROR.ordinal()][index] = applyValue(count > 0 ? 1 : 0, 1, (float)mse);
            }
            if (result.data[OutputVariant.ROOT_MEAN_SQUARED_ERROR.ordinal()] != null) {
                result.data[OutputVariant.ROOT_MEAN_SQUARED_ERROR.ordinal()][index] = applyValue(count > 0 ? 1 : 0, 1, (float)Math.sqrt(mse));
            }
        }
        
    }
    
    private static class SingleLowerBoundEstimatorEvaluator implements ImagePositionEvaluator {

        private Point2d[] anchors;
        private ErrorModel errorModel;
        private BasicLaterationAlgorithm algorithm;
        
        public SingleLowerBoundEstimatorEvaluator(BasicLaterationAlgorithm algorithm,
                Point2d[] anchors, ErrorModel errorModel) {
            this.anchors = anchors;
            this.errorModel = errorModel;
            this.algorithm = algorithm;
        }
        
        public void evaluatePosition(Point2d position, double[] distances, int index, LS2Result result) {
            int count = 0;
            double error = 0;
            Point2d estimate = algorithm.localize(anchors, distances, position, errorModel, result.width, result.height);
            if (estimate != null) {
                count++;
                // Lower bound of error is stored in both point coordinates
                error = estimate.x;
            }
            // Write results back at given position
            if (result.data[OutputVariant.AVERAGE_ERROR.ordinal()] != null) {
                result.data[OutputVariant.AVERAGE_ERROR.ordinal()][index] = LS2Result.ERROR;
            }
            if (result.data[OutputVariant.MAXIMUM_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MAXIMUM_ERROR.ordinal()][index] = LS2Result.ERROR;
            }
            if (result.data[OutputVariant.MINIMUM_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MINIMUM_ERROR.ordinal()][index] = LS2Result.ERROR;
            }
            if (result.data[OutputVariant.MEAN_SQUARED_ERROR.ordinal()] != null) {
                result.data[OutputVariant.MEAN_SQUARED_ERROR.ordinal()][index] = applyValue(count, 1, (float)error);
            }
            if (result.data[OutputVariant.ROOT_MEAN_SQUARED_ERROR.ordinal()] != null) {
                result.data[OutputVariant.ROOT_MEAN_SQUARED_ERROR.ordinal()][index] = applyValue(count, 1, (float)Math.sqrt(error));
            }
        }
        
    }
    
    private static float applyValue(int count, int needed, float value) {
        if (count >= needed) {
            return value / count;
        } else {
            return LS2Result.ERROR;
        }
    }
    
}
